@@ -1,17 +1,24 @@
 """Test GatewayClient"""
+import asyncio
 import json
+import logging
 import os
 import uuid
 from datetime import datetime
 from io import BytesIO
-from unittest.mock import patch
+from queue import Empty
+from unittest.mock import MagicMock, patch
 
 import pytest
 import tornado
 from tornado.httpclient import HTTPRequest, HTTPResponse
 from tornado.web import HTTPError
 
-from jupyter_server.gateway.managers import GatewayClient
+from jupyter_server.gateway.managers import (
+    ChannelQueue,
+    GatewayClient,
+    GatewayKernelManager,
+)
 from jupyter_server.utils import ensure_async
 
 from .utils import expected_http_error
@@ -29,7 +36,11 @@ def generate_kernelspec(name):
             "metadata": {},
         }
     }
-    kernelspec_stanza = {"name": name, "spec": spec_stanza, "resources": {}}
+    kernelspec_stanza = {
+        "name": name,
+        "spec": spec_stanza,
+        "resources": {"logo-64x64": f"f/kernelspecs/{name}/logo-64x64.png"},
+    }
     return kernelspec_stanza
 
 
@@ -161,6 +172,15 @@ mock_gateway_url = "http://mock-gateway-server:8889"
 mock_http_user = "alice"
 
 
+def mock_websocket_create_connection(recv_side_effect=None):
+    def helper(*args, **kwargs):
+        mock = MagicMock()
+        mock.recv = MagicMock(side_effect=recv_side_effect)
+        return mock
+
+    return helper
+
+
 @pytest.fixture
 def init_gateway(monkeypatch):
     """Initializes the server for use as a gateway client."""
@@ -170,6 +190,7 @@ def init_gateway(monkeypatch):
     monkeypatch.setenv("JUPYTER_GATEWAY_HTTP_USER", mock_http_user)
     monkeypatch.setenv("JUPYTER_GATEWAY_REQUEST_TIMEOUT", "44.4")
     monkeypatch.setenv("JUPYTER_GATEWAY_CONNECT_TIMEOUT", "44.4")
+    monkeypatch.setenv("JUPYTER_GATEWAY_LAUNCH_TIMEOUT_PAD", "1.1")
     yield
     GatewayClient.clear_instance()
 
@@ -182,11 +203,10 @@ async def test_gateway_env_options(init_gateway, jp_serverapp):
         jp_serverapp.gateway_config.connect_timeout == jp_serverapp.gateway_config.request_timeout
     )
     assert jp_serverapp.gateway_config.connect_timeout == 44.4
+    assert jp_serverapp.gateway_config.launch_timeout_pad == 1.1
 
     GatewayClient.instance().init_static_args()
-    assert GatewayClient.instance().KERNEL_LAUNCH_TIMEOUT == int(
-        jp_serverapp.gateway_config.request_timeout
-    )
+    assert GatewayClient.instance().KERNEL_LAUNCH_TIMEOUT == 43
 
 
 async def test_gateway_cli_options(jp_configurable_serverapp):
@@ -195,6 +215,7 @@ async def test_gateway_cli_options(jp_configurable_serverapp):
         "--GatewayClient.http_user=" + mock_http_user,
         "--GatewayClient.connect_timeout=44.4",
         "--GatewayClient.request_timeout=96.0",
+        "--GatewayClient.launch_timeout_pad=5.1",
     ]
 
     GatewayClient.clear_instance()
@@ -205,10 +226,40 @@ async def test_gateway_cli_options(jp_configurable_serverapp):
     assert app.gateway_config.http_user == mock_http_user
     assert app.gateway_config.connect_timeout == 44.4
     assert app.gateway_config.request_timeout == 96.0
+    assert app.gateway_config.launch_timeout_pad == 5.1
     GatewayClient.instance().init_static_args()
     assert (
-        GatewayClient.instance().KERNEL_LAUNCH_TIMEOUT == 96
-    )  # Ensure KLT gets set from request-timeout
+        GatewayClient.instance().KERNEL_LAUNCH_TIMEOUT == 90
+    )  # Ensure KLT gets set from request-timeout - launch_timeout_pad
+    GatewayClient.clear_instance()
+
+
+@pytest.mark.parametrize(
+    "request_timeout,kernel_launch_timeout,expected_request_timeout,expected_kernel_launch_timeout",
+    [(50, 10, 50, 45), (10, 50, 55, 50)],
+)
+async def test_gateway_request_timeout_pad_option(
+    jp_configurable_serverapp,
+    monkeypatch,
+    request_timeout,
+    kernel_launch_timeout,
+    expected_request_timeout,
+    expected_kernel_launch_timeout,
+):
+    argv = [
+        f"--GatewayClient.request_timeout={request_timeout}",
+        "--GatewayClient.launch_timeout_pad=5",
+    ]
+
+    GatewayClient.clear_instance()
+    app = jp_configurable_serverapp(argv=argv)
+
+    monkeypatch.setattr(GatewayClient, "KERNEL_LAUNCH_TIMEOUT", kernel_launch_timeout)
+    GatewayClient.instance().init_static_args()
+
+    assert app.gateway_config.request_timeout == expected_request_timeout
+    assert GatewayClient.instance().KERNEL_LAUNCH_TIMEOUT == expected_kernel_launch_timeout
+
     GatewayClient.clear_instance()
 
 
@@ -228,6 +279,9 @@ async def test_gateway_get_kernelspecs(init_gateway, jp_fetch):
         kspecs = content.get("kernelspecs")
         assert len(kspecs) == 2
         assert kspecs.get("kspec_bar").get("name") == "kspec_bar"
+        assert (
+            kspecs.get("kspec_bar").get("resources")["logo-64x64"].startswith(jp_serverapp.base_url)
+        )
 
 
 async def test_gateway_get_named_kernelspec(init_gateway, jp_fetch):
@@ -318,6 +372,78 @@ async def test_gateway_shutdown(init_gateway, jp_serverapp, jp_fetch, missing_ke
     assert await is_kernel_running(jp_fetch, k2) is False
 
 
+@patch("websocket.create_connection", mock_websocket_create_connection(recv_side_effect=Exception))
+async def test_kernel_client_response_router_notifies_channel_queue_when_finished(
+    init_gateway, jp_serverapp, jp_fetch
+):
+    # create
+    kernel_id = await create_kernel(jp_fetch, "kspec_bar")
+
+    # get kernel manager
+    km: GatewayKernelManager = jp_serverapp.kernel_manager.get_kernel(kernel_id)
+
+    # create kernel client
+    kc = km.client()
+
+    await ensure_async(kc.start_channels())
+
+    with pytest.raises(RuntimeError):
+        await kc.iopub_channel.get_msg(timeout=10)
+
+    all_channels = [
+        kc.shell_channel,
+        kc.iopub_channel,
+        kc.stdin_channel,
+        kc.hb_channel,
+        kc.control_channel,
+    ]
+    assert all(channel.response_router_finished if True else False for channel in all_channels)
+
+    await ensure_async(kc.stop_channels())
+
+    # delete
+    await delete_kernel(jp_fetch, kernel_id)
+
+
+async def test_channel_queue_get_msg_with_invalid_timeout():
+    queue = ChannelQueue("iopub", MagicMock(), logging.getLogger())
+
+    with pytest.raises(ValueError):
+        await queue.get_msg(timeout=-1)
+
+
+async def test_channel_queue_get_msg_raises_empty_after_timeout():
+    queue = ChannelQueue("iopub", MagicMock(), logging.getLogger())
+
+    with pytest.raises(Empty):
+        await asyncio.wait_for(queue.get_msg(timeout=0.1), 2)
+
+
+async def test_channel_queue_get_msg_without_timeout():
+    queue = ChannelQueue("iopub", MagicMock(), logging.getLogger())
+
+    with pytest.raises(asyncio.TimeoutError):
+        await asyncio.wait_for(queue.get_msg(timeout=None), 1)
+
+
+async def test_channel_queue_get_msg_with_existing_item():
+    sent_message = {"msg_id": 1, "msg_type": 2}
+    queue = ChannelQueue("iopub", MagicMock(), logging.getLogger())
+    queue.put_nowait(sent_message)
+
+    received_message = await asyncio.wait_for(queue.get_msg(timeout=None), 1)
+
+    assert received_message == sent_message
+
+
+async def test_channel_queue_get_msg_when_response_router_had_finished():
+    queue = ChannelQueue("iopub", MagicMock(), logging.getLogger())
+    queue.response_router_finished = True
+
+    with pytest.raises(RuntimeError):
+        await queue.get_msg()
+
+
 #
 # Test methods below...
 #
